unit integral;

interface

const
	max_dim = 10;
	max_deg = 96;

type 
	real_fun = function(x : real) : real;
	real_fun2 = function(x, y : real) : real;
	real_vec = array[1..max_dim + 1] of real;
	index = array[1..max_dim + 1] of word;
	vec_fun = function(j : word;  x : real_vec) : real;

var 
	no_evaluations, highest_level : word;

function simpson(F : real_fun;  x0, x1 : real; div_no : word) : real;
function double_simpson(F : real_fun2; x0, x1, y0, y1 : real;
		x_div, y_div : word) : real;
function adaptive_simpson(F : real_fun; x0, x1, eps, eta : real) : real;
function romberg(f : real_fun; x0, x1, eps, eta : real;
		min, max : word) : real;
function gauss3(F : real_fun; x0, x1 : real;  n : word): real;
procedure compute_gauss_coeffs(deg : word);
function gauss(F : real_fun; x0, x1 : real;  deg : word) : real;

implementation

uses  crt;

var
	zero, weight : array[1..max_deg] of real;

function simpson(F : real_fun;  x0, x1 : real; div_no : word) : real;
var
	x, dx, sum : real;
	j : word;
begin
	dx := (x1 - x0) / (2.0 * div_no);
	sum := F(x0) + F(x1);
	x := x0;
	for j := 1 to 2 * div_no - 1  do
	begin
		x := x + dx;
		if odd(j) then
			sum := sum + 4.0 * F(x)
		else
			sum := sum + 2.0 * F(x);
	end;
	simpson := dx * sum / 3.0;
end;

function double_simpson(F : real_fun2; x0, x1, y0, y1 : real;
		x_div, y_div : word) : real;
var
	dx, dy, x, sum : real;
	i : word;

function simple_simpson(x : real) : real;
var
	y, sum : real;
	j, v : word;
begin
	sum := F(x, y0) + F(x, y1);
	y := y0;
	for  j := 1  to  2 * y_div - 1  do
	begin
		y := y + dy;
		if  odd(j)  then
			sum := sum + 4.0 * F(x, y)
		else  
			sum := sum + 2.0 * F(x, y);
	end;
	simple_simpson := sum;
end;{simple_simpson}

begin{double_simpson}
	dx := (x1 - x0) / (2.0 * x_div);
	dy := (y1 - y0) / (2.0 * y_div);
	x := x0;
	sum := simple_simpson(x0) + simple_simpson(x1);
	for i := 1 to 2 * x_div - 1  do
	begin
		x := x + dx;
		if odd(i) then
			sum := sum + 4.0 * simple_simpson(x)
		else
			sum := sum + 2.0 * simple_simpson(x);
	end;
	double_simpson := dx * dy * sum / 9.0;
end;{double_simpson}

function adaptive_simpson(F : real_fun; x0, x1, eps, eta : real) : real;
const
	max_level = 35;
var
	k, nest_level : word;
	integral_abs : real;

function simpson3point(x0, delta_x, estimate, integral_abs,
		eps, eta, left, middle, right : real) : real;
var
	dx3, sum, eps3, eta3, factor, left_integ,
	middle_integ, right_integ, F1, F2, F4, F5 : real;
begin
	inc(nest_level);
	dx3 := delta_x / 3.0;
	F1 := F(x0 + 0.5 * dx3);
	F2 := F(x0 + dx3);
	F4 := F(x0 + 2.0 * dx3);
	F5 := F(x0 + 2.5 * dx3);
	inc(no_evaluations, 4);
	factor := dx3 / 6.0;
	left_integ := factor * (left + 4.0 * F1 + F2);
	middle_integ := factor * (F2 + 4.0 * middle + F4);
	right_integ := factor * (F4 + 4.0 * F5 + right);
	sum := left_integ + middle_integ + right_integ;
	integral_abs := integral_abs - abs(estimate)
		+ abs(left_integ) + abs(middle_integ) + abs(right_integ);
	if (nest_level > 1 ) and ((nest_level = max_level) or
		(abs(sum - estimate) <= eps + eta * integral_abs)) then
			simpson3point := sum
	else
	begin
		if nest_level > highest_level then 
			inc(highest_level);
		eps3 := 0.577 * eps;
		eta3 := 0.577 * eta;
		left_integ := simpson3point(x0, dx3, left_integ, integral_abs,
			eps3, eta3, left, F1, F2);
		middle_integ := simpson3point(x0 + dx3, dx3,
			middle_integ, integral_abs, eps3, eta3, F2, middle, F4);
		right_integ := simpson3point(x0 + 2.0 * dx3, dx3,
			right_integ, integral_abs, eps3, eta3, F4, F5, right);
		simpson3point := left_integ + middle_integ + right_integ;
	end;
	dec(nest_level);
end;{simpson3point}

begin{adaptive_simpson}
	nest_level := 1;
	highest_level := 1;
	no_evaluations := 3;
	adaptive_simpson := simpson3point(x0, x1 - x0, 0.0, 0.0, eps, eta,
		F(x0), F(x0 + 0.5 * (x1 - x0)), F(x1));
end;{adaptive_simpson}

function romberg(F : real_fun; x0, x1, eps, eta : real; min, max : word) : real;
const
	abs_max = 30;
var
	p, dx, error, F_of_x0, F_of_x1, F_of_xk,
	roundoff_error, integral_abs, tolerance,
	previous_estimate, current_estimate,
	mid_sum, temp_sum, mid_sum_abs : real;
	table : array[0..abs_max] of real;
	j, n : word;
	k, r : longint;
	done : boolean;
	denom : array[1..abs_max] of real;
begin
	p := 1.0;
	for k := 1 to abs_max do
	begin
		p := 4.0 * p;  
		denom[k] := 1.0 / (p - 1.0);
	end;
	dx := x1 - x0;
	F_of_x0 := F(x0);  
	F_of_x1 := F(x1);
	current_estimate := 0.0;
	previous_estimate := 0.0;
	done := false;
	table[0] := 0.5 * dx * (F_of_x0 + F_of_x1);
	integral_abs := 0.5 * abs(dx) * (abs(F_of_x0) + abs(F_of_x1));
	n := 1;
	r := 1;
	repeat
		dx := 0.5 * dx;
		mid_sum := 0.0;
		mid_sum_abs := 0.0;
		roundoff_error := 0.0;
		for k := 1 to r do
		begin
			F_of_xk := F(x0 + (2 * k - 1) * dx);
			mid_sum_abs := mid_sum_abs + abs(F_of_xk);
			F_of_xk := F_of_xk + roundoff_error;
			temp_sum := mid_sum + F_of_xk;
			roundoff_error := (mid_sum - temp_sum) + F_of_xk;
			mid_sum := temp_sum;
			if keypressed then
				halt;
		end;
		table[n] := 0.5 * table[n - 1] + dx * mid_sum;
		integral_abs := 0.5 * integral_abs + abs(dx) * mid_sum_abs;
		for j := n - 1 downto 0 do 
			table[j] := table[j + 1] + denom[n - j]*(table[j + 1] - table[j]);
		if n >= min then
		begin
			tolerance := eta * integral_abs + eps;
			error := abs(table[0] - current_estimate)
				+ abs(current_estimate - previous_estimate);
			done := (error < tolerance);
		end;
		inc(n);
		done := done or (n > max);
		previous_estimate := current_estimate;
		current_estimate := table[0];
		r := r + r;
	until done;
	romberg := current_estimate;
end;

function gauss3(F : real_fun; x0, x1 : real;  n : word) : real;
var
	t, sum, x, z, dx : real;
	i, k : word;
	gzero, gweight : array[1..3] of real;

procedure  initialize_constants;
var
	s, t : real;
	j : word;
begin
	gzero[1] := -sqrt(0.6);
	gzero[2] := 0.0;
	gzero[3] := sqrt(0.6);
	gweight[1] := 5.0 / 9.0;
	gweight[2] := 8.0 / 9.0;
	gweight[3] := 5.0 / 9.0;
	for j := 1 to 3 do
	begin
		gzero[j] := 0.5 * (1.0 + gzero[j]);
		gweight[j] := 0.5 * gweight[j];
	end;
end;{initialize_constants}

begin{gauss3}
	initialize_constants;
	dx := (x1 - x0) / n;
	x := x0;
	sum := 0.0;
	for  i := 0  to  n - 1  do
	begin
		t := 0.0;
		for k := 1 to 3 do
		begin
			z := x + dx * gzero[k];
			t := t + gweight[k] * F(z)
		end;
		sum := sum + dx * t;
		x := x + dx;
	end;
	gauss3 := sum;
end;{gauss3}

procedure compute_gauss_coeffs(deg : word);
const
	eps = 6.0e-20;
var
	i, index : word;
	P0k, P0k_1, D0k, P1k, P1k_1, D1k,
	x0, x1, y, z, dx, x, u : real;

procedure legendre_poly(n : word;  x : real; var  Pk, Pk_1, Dk : real);
var
	Pk_2, Dk_1, Dk_2 : real;
	i, j, k : word;
begin
	if n = 0 then
	begin  
		Pk := 1.0;
		Dk := 0.0;
	end
	else
	begin
		Pk_1 := 1.0;
		Pk := x;
		Dk_1 := 0.0;
		Dk := 1.0;
		i := 3;
		j := 1;
		for k := 2 to n do
		begin
			Pk_2 := Pk_1;
			Pk_1 := Pk;
			Dk_2 := Dk_1;
			Dk_1 := Dk;
			Pk := (i * x * Pk_1 - j * Pk_2 ) / k;
			Dk := (i * (Pk_1 + x * Dk_1) - j * Dk_2 ) / k;
			inc(i, 2);
			inc(j);
		end;
	end
end;{legendre_poly}

begin{compute_gauss_coeffs}
	index := (deg + 1) div 2;
	dx := 1.0 / (10.0 * deg);
	x0 := 0.0;
	x1 := x0 + dx;
	if  odd(deg)  then
	begin
		zero[index] := 0.0;
		legendre_poly(deg, x0, P0k, P0k_1, D0k);
		weight[index] := 2.0 / (P0k_1 * D0k * deg);
	end;
	for  i := 0  to  10*deg - 1  do
	begin
		x0 := x1;
		x1 := x1 + dx;
		legendre_poly(deg, x0, P0k, P0k_1, D0k);
		legendre_poly(deg, x1, P1k, P1k_1, D1k);
		if  P0k * P1k <= 0.0  then
		begin
			x := x0 - P0k * dx / (P1k - P0k);
			legendre_poly(deg, x, P0k, P0k_1, D0k);
			u := P0k / D0k;
			y := x - u;
			while  abs(x - y) >= eps  do
			begin
				if  keypressed  then
				begin
					writeln(' >= eps loop:  ', x:10:10,'   ', 
						y:10:10, '   ', abs(x-y):10);
					readln;
				end;
				x := y;
				legendre_poly(deg, x, P0k, P0k_1, D0k);
				u := P0k / D0k;
				y := x - u;
			end;
			inc(index);
			legendre_poly(deg, y, P0k, P0k_1, D0k);
			zero[index] := y;
			weight[index] := 2.0 / (P0k_1 * D0k * deg);
			if  index = deg  then  
				break;
		end;
	end;
	for  i := 1  to  deg  div 2  do
	begin
		zero[i] := - zero[deg - i + 1];
		weight[i] := weight[deg - i + 1];
	end;
end;{compute_gauss_coeffs}

function gauss(F : real_fun; x0, x1 : real;  deg : word) : real;
var
	index : word;
	a, b, sum : real;
begin
	a := 0.5 * (x1 - x0);
	b := 0.5 * (x1 + x0);
	sum := 0.0;
	for  index := 1  to  deg  do
	begin
		sum := sum + F(a * zero[index] + b) * weight[index];
		if keypressed then  
			halt;
	end;
	gauss := a * sum;
end;


end.



	