unit rkutta;

interface

const
	max_dim = 10;

type
	vector = array[1..max_dim] of real;
	derivative = procedure(const  t : real; const  x : vector;
		var  w : vector);

function max(a, b : real) : real;
function sup_norm(const  dim : word; const  z : vector): real;
function x2y(x, y : real) : real;
procedure Euler(dim : word; F : derivative; const  t0, dt : real; 
		const  x0 : vector; var  x1, y1, error : vector);
procedure KM34(dim : word; F : derivative; const  t0, dt : real;
		const  x0 : vector; var  x1, y1, error : vector);
procedure RKN34(dim : word; F : derivative; const  t0, dt : real;
		const  x0 : vector; var  x1, y1, error : vector);
procedure RK4(dim : word; F : derivative; const  t0, dt : real;
		const  x0 : vector; var  x1, y1, error : vector);
procedure RKF45(dim : word; F : derivative; const  t0, dt : real;
		const  x0 : vector; var  x1, y1, error : vector);

implementation

function max(a, b : real) : real;
begin
	if a >= b then
		max := a
	else
		max := b;
end;

function sup_norm(const  dim : word; const  z : vector) : real;
var
	k : word;
	w : real;
begin
	w := 0.0;
	for k := 1 to dim do
		w := max(w, abs(z[k]));
	sup_norm := w;
end;

function x2y(x, y : real) : real;
begin
	x2y := exp(y * ln(x));
end;

procedure Euler(dim : word; F : derivative; const  t0, dt : real; 
		const  x0 : vector; var  x1, y1, error : vector);
var
	slope0, slope1, z, w : vector;
	j : word;
begin
	F(t0, x0, slope0);
	for j := 1 to dim do
		z[j] := x0[j] + dt * slope0[j];
	F(t0 + dt, z, slope1);
	for j := 1 to dim do
		x1[j] := x0[j] + 0.5 * dt * (slope0[j] + slope1[j]);
	for j := 1 to dim do
		z[j] := x0[j] + 0.5 * dt * slope0[j];
	F(t0 + 0.5 * dt, z, slope1);
	for j := 1 to dim do
		w[j] := x0[j] + 0.25 * dt * (slope0[j] + slope1[j]);
	F(t0 + 0.5 * dt, w, slope0);
	for j := 1 to dim do
		z[j] := w[j] + 0.5 * dt * slope0[j];
	F(t0 + dt, z, slope1);
	for j := 1 to dim do
	begin
		y1[j] := w[j] + 0.25 * dt * (slope0[j] + slope1[j]);
		error[j] := x1[j] - y1[j];
	end;
end;

procedure KM34(dim : word; F : derivative; const  t0, dt : real; 
		const  x0 : vector; var  x1, y1, error : vector);
const
	bound = 4;
	abscissa : array[1..bound] of real = (1./3., 1./3., 1./2., 1. );
	weight : array[1..bound, 0..bound - 1] of real = ( 
		(1./3., 0., 0., 0. ),
		(1./6., 1./6., 0., 0. ),
		(1./8., 0., 3./8., 0. ),
		(1./2., 0., -3./2., 2. ) );
	mult1 : array[0..bound] of real = 
		(1./2., 0., -3./2., 2., 0. );
	mult2 : array[0..bound] of real = 
		(1./6., 0., 0., 2./3., 1./6. );
	error_coeff : array[0..bound] of real = 
		(1./3., 0., -3./2., 4./3., -1./6. );

var
	slope, z : array[0..bound] of vector;
	s, s1, s2 : real;
	i, j, k : word;

begin{KM34}
	F(t0, x0, slope[0]);
	for  i := 1  to  bound  do
	begin
		for  j := 1  to  dim  do
		begin
			s := 0.0;
			for  k := 0  to  i - 1  do
				s := s + weight[i, k] * slope[k, j];
			z[i, j] := x0[j] + dt * s;
		end;
		F(t0 + abscissa[i] * dt, z[i], slope[i]);
	end;
	for  j := 1  to  dim  do
	begin
		s1 := 0.0;
		s2 := 0.0;
		for  k := 0  to  bound  do
		begin
			s1 := s1 + mult1[k] * slope[k, j];
			s2 := s2 + mult2[k] * slope[k, j];
		end;
		x1[j] := x0[j] + dt * s1;
		y1[j] := x0[j] + dt * s2;
		error[j] := x1[j] - y1[j];
	end;
end;{KM34}

procedure RKN34(dim : word; F : derivative; const  t0, dt : real;
		const  x0 : vector; var  x1, y1, error : vector);
const
	bound = 4;
	abscissa : array[1..bound] of real = 
		(3./8., 9./16., 25./32., 1. );
	weight : array[1..bound, 0..bound - 1] of real = ( 
		(3./8., 0., 0., 0. ),
		(0., 9./16., 0., 0. ),
		(-125./672., 325./336., 0., 0. ),
		(371./891., -200./297., 1120./891., 0. ) );
	mult1 : array[0..bound] of real = 
		(37./225., 44./117., 0., 448./975., 0. );
	mult2 : array[0..bound] of real = 
		(25./162., 32./135., 256./567., 0., 11./70. );
	error_coeff : array[0..bound] of real = 
		(41./4050., 244./1755., -256./567., 448./975., -11./70. );

var
	z, slope : array[0..bound] of vector;
	s, s1, s2 : real;
	i, j, k : word;

begin{RKN34}
	F(t0, x0, slope[0]);
	for  i := 1  to  bound  do
	begin
		for  j := 1  to  dim  do
		begin
			s := 0.0;
			for  k := 0  to  i - 1  do
				s := s + weight[i, k] * slope[k, j];
			z[i, j] := x0[j] + dt * s;
		end;
		F(t0 + abscissa[i] * dt, z[i], slope[i]);
	end;
	for  j := 1  to  dim  do
	begin
		s1 := 0.0;
		s2 := 0.0;
		for  k := 0  to  bound  do
		begin
			s1 := s1 + mult1[k] * slope[k, j];
			s2 := s2 + mult2[k] * slope[k, j];
		end;
		x1[j] := x0[j] + dt * s1;
		y1[j] := x0[j] + dt * s2;
		error[j] := x1[j] - y1[j];
	end;
end;{RKN34}

procedure RK4(dim : word; F : derivative; const  t0, dt : real; 
		const  x0 : vector; var  x1, y1, error : vector);
const
	bound = 3;
	abscissa : array[1..bound] of real = (1./2., 1./2., 1. );
	weight : array[1..bound, 0..bound - 1] of real = ( 
		(1./2., 0., 0. ), 
		(0.0, 1./2., 0. ),
		(0., 0., 1. ) );
	mult1 : array[0..bound] of real = (1./6., 1./3., 1./3., 1./6. );
var
	slope : array[0..bound] of vector;
	w : vector;
	s, s1, s2 : real;
	j : word;

procedure  step(const  t0, dt : real; const  x0 : vector; var  y : vector);
var
	i, j, k : word;
	z : array[1..bound] of vector;
begin
	F(t0, x0, slope[0]);
	for  i := 1  to  bound  do
	begin
		for  j := 1  to  dim  do
		begin
			s := 0.0;
			for  k := 0  to  i - 1  do
				s := s + weight[i, k] * slope[k, j];
			z[i, j] := x0[j] + dt * s;
		end;
		F(t0 + abscissa[i] * dt, z[i], slope[i]);
	end;
	for  j := 1  to  dim  do
	begin
		s1 := 0.0;
		for  k := 0  to  bound  do
			s1 := s1 + mult1[k] * slope[k, j];
		y[j] := x0[j] + dt * s1;
	end;
end;{step}

begin{RK4}
	step(t0, dt, x0, x1);
	step(t0, 0.5 * dt, x0, w);
	step(t0 + 0.5*dt, 0.5 * dt, w, y1);
	for  j := 1  to  dim  do
		error[j] := x1[j] - y1[j];
end;{RK4}

procedure RKF45(dim : word; F : derivative; const  t0, dt : real; 
		const  x0 : vector; var  x1, y1, error : vector);
const
	bound = 5;
	abscissa : array[1..bound] of real = 
		(1./4., 3./8., 12./13., 1., 1./2. );
	weight : array[1..bound, 0..bound - 1] of real = ( 
		(1./4., 0., 0., 0., 0. ),
		(3./32., 9./32., 0., 0., 0. ),
		(1932./2197., -7200./2197,  7296./2197., 0., 0. ),
		(439./216., -8., 3680./513., -845./4104., 0. ),
		(-8./27., 2., -3544./2565., 1859./4104., -11./40. ) );
	mult1 : array[0..bound] of real = 
		(25./216., 0., 1408./2565., 2197./4104., -1./5., 0 );
	mult2 : array[0..bound] of real = 
		(16./135., 0., 6656./12825., 28561./56430., -9./50., 2./55. );
	error_coeff : array[0..bound] of real = 
		(-1./360., 0., 128./4275., 2197./75240., -1./50., -2./55. );
var
	slope : array[0..bound] of vector;
	z : array[1..bound] of vector;
	s, s1, s2 : real;
	i, j, k : word;

begin{RKF45}
	F(t0, x0, slope[0]);
	for  i := 1  to  bound  do
	begin
		for  j := 1  to  dim  do
		begin
			s := 0.0;
			for  k := 0  to  i - 1  do
				s := s + weight[i, k] * slope[k, j];
			z[i, j] := x0[j] + dt * s;
		end;
		F(t0 + abscissa[i] * dt, z[i], slope[i]);
	end;
	for  j := 1  to  dim  do
	begin
		s1 := 0.0;
		s2 := 0.0;
		for  k := 0  to  bound  do
		begin
			s1 := s1 + mult1[k] * slope[k, j];
			s2 := s2 + mult2[k] * slope[k, j];
		end;
		x1[j] := x0[j] + dt * s1;
		y1[j] := x0[j] + dt * s2;
		error[j] := x1[j] - y1[j];
	end;
end;{RKF45}

end.
	